import { Ref, ComputedRef, VNodeProps, defineComponent, ComponentPublicInstance, UnwrapRef } from 'vue';

/**
 * Value may or may not be a `ref`.
 *
 * ```ts
 * type MaybeRef<T> = T | Ref<T>
 * ```
 */
type MaybeRef<T> = T | Ref<T> | ComputedRef<T>;
/**
 * May be a simple ref (nor computed ref)
 */
type MaybeBaseRef<T> = T | Ref<T>;
/** VNodeProps Object */
interface MergedVNodeProps extends VNodeProps {
    ref: TemplateRef | ((el: TemplateRef | null) => void);
}
/**
 * Any function
 */
type Fn = () => void;
/**
 * Any function
 */
type AnyFn = (...args: any[]) => any;
/**
 * A ref that allow to set null or undefined
 */
type RemovableRef<T> = Omit<Ref<T>, "value"> & {
    get value(): T;
    set value(value: T | null | undefined);
};
/**
 * @deprecated Use `RemovableRef`
 */
type RemoveableRef<T> = RemovableRef<T>;
/**
 * Maybe it's a ref, or a plain value, or a getter function
 *
 * ```ts
 * type MaybeComputedRef<T> = (() => T) | T | Ref<T> | ComputedRef<T>
 * ```
 */
type MaybeComputedRef<T> = MaybeReadonlyRef<T> | MaybeRef<T>;
/**
 * Maybe it's a computed ref, or a getter function
 *
 * ```ts
 * type MaybeReadonlyRef<T> = (() => T) | ComputedRef<T>
 * ```
 */
type MaybeReadonlyRef<T> = (() => T) | ComputedRef<T>;
/**
 * Make all the nested attributes of an object or array to MaybeRef<T>
 *
 * Good for accepting options that will be wrapped with `reactive` or `ref`
 *
 * ```ts
 * UnwrapRef<DeepMaybeRef<T>> === T
 * ```
 */
type DeepMaybeRef<T> = T extends Ref<infer V> ? MaybeRef<V> : T extends Array<any> | object ? {
    [K in keyof T]: DeepMaybeRef<T[K]>;
} : MaybeRef<T>;
type Arrayable<T> = T[] | T;
/**
 * Assign property types from right to left.
 * Handy for overriding e.g. `onChange` from an HTMLElement with your own type
 */
type Assign<Target, Source> = Omit<Target, keyof Source> & Source;
/**
 * Type to make properties optional and preserve their type
 */
type Optional<T, K extends keyof T> = Pick<Partial<T>, K> & Omit<T, K>;

/**
 * Interface for node provided by template ref
 */
type TemplateRef = Element | VueComponentInstance | undefined | null;
/**
 * For internal use
 *
 * Creates refs that will be bound to the template/render function.
 *
 * Why not just use the regular `ref(null)` and bind it to the element?
 *
 * 1. To avoid unwrapping template refs which maybe components. This hook will always
 *    give us the actual element being bound the the element, and not the component
 *    options.
 *
 * 2. In some cases where we need an up-to-date value of the ref node,
 *    from the consuming component, we can use this hook.
 *
 * @returns []
 */
declare function useRef<T extends HTMLElement>(): [
    (el: TemplateRef | null) => void,
    Ref<T | null>
];
/** Vue Component HTML Element Instance */
type VueComponentInstance = InstanceType<ReturnType<typeof defineComponent>>;
/** Ref may or may not be an HTML Element or VueComponent instance */
type MaybeElementRef = MaybeRef<Element | ComponentPublicInstance | VueComponentInstance | undefined | null>;
/**
 * Unwraps element from ref
 * @param elementRef Ref of template node
 */
declare function unrefElement(elementRef: MaybeElementRef): UnwrapRef<MaybeElementRef>;
/**
 * Creates a ref whose value updates are debounced
 *
 * @example Simple example
 *
 * ```ts
 * const foo = useDebouncedRef('bar')
 * foo.value = 'baz'
 *
 * // foo.value to be updated to 'baz' after the delay of 300ms
 * ```
 *
 * @example Custom delay
 *
 * ```ts
 * const foo = useDebouncedRef('bar', 500)
 * foo.value = 'baz'
 *
 * // foo.value to be updated to 'baz' after the delay of 500ms
 * ```
 */
declare function useDebouncedRef<T = unknown>(initialValue: T, delay?: number, immediate?: boolean): Ref<UnwrapRef<T>>;
type DebouncedRef = Ref<unknown>;
declare function contains(containers: Set<HTMLElement>, element: HTMLElement): boolean;
interface ConfigurableWindow {
    window?: Window;
}
interface ConfigurableDocument {
    document?: Document;
}
interface ConfigurableDocumentOrShadowRoot {
    document?: DocumentOrShadowRoot;
}
interface ConfigurableNavigator {
    navigator?: Navigator;
}
interface ConfigurableLocation {
    location?: Location;
}
declare const defaultWindow: (Window & typeof globalThis) | undefined;
declare const defaultDocument: Document | undefined;
declare const defaultNavigator: Navigator | undefined;
declare const defaultLocation: Location | undefined;
/**
 * Sorts an array of nodes by their position in the DOM
 * @param nodes
 * @param resolveKey
 * @returns
 */
declare function sortByDomNode<T>(nodes: T[], resolveKey?: (item: T) => HTMLElement | null): T[];

export { AnyFn as A, ConfigurableDocument as C, DeepMaybeRef as D, Fn as F, MaybeBaseRef as M, Optional as O, RemovableRef as R, TemplateRef as T, VueComponentInstance as V, Arrayable as a, MaybeComputedRef as b, MaybeReadonlyRef as c, MaybeRef as d, RemoveableRef as e, contains as f, DebouncedRef as g, defaultDocument as h, defaultLocation as i, defaultNavigator as j, defaultWindow as k, MaybeElementRef as l, useDebouncedRef as m, useRef as n, ConfigurableDocumentOrShadowRoot as o, ConfigurableLocation as p, ConfigurableNavigator as q, ConfigurableWindow as r, sortByDomNode as s, MergedVNodeProps as t, unrefElement as u, Assign as v };
